"""Report analyzers output to console"""
import os
from typing import Dict, List, Tuple

import humanize
import numpy as np
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree

from neatmesh.analyzer import Analyzer2D, Analyzer3D
from neatmesh.reader import MeshReader


class Reporter:
    """Base class for 2D & 3D reporters"""

    def __init__(self, console: Console, mesh: MeshReader, filename: str) -> None:
        self.console = console
        self.mesh = mesh
        self.filename = filename
        self.concerns: List[str] = []

    @staticmethod
    def stats_from_array(array: np.ndarray) -> Tuple[float, ...]:
        """Calculate maximum, minimum, mean and standard deviation for a
        given numpy array

        Args:
            array (np.ndarray): Input array of shape (n,)

        Returns:
            Tuple[float, ...]: max., min., mean and std.
        """
        arr_max = np.nanmax(array)
        arr_min = np.nanmin(array)
        arr_mean = np.nanmean(array)
        arr_std = np.nanstd(array)

        return arr_max, arr_min, arr_mean, arr_std

    def report_mesh_stats(self, quality_metrics_dict):
        self.console.print()

        stats_table = Table(box=box.SIMPLE)
        stats_table.add_column("", justify="left")
        stats_table.add_column("Max.", justify="right")
        stats_table.add_column("Min.", justify="right")
        stats_table.add_column("Mean", justify="right")
        stats_table.add_column("Std.", justify="right")
        stats_table.add_column("", justify="right")

        for metric_name, metric_dict in quality_metrics_dict.items():
            # Avoid calling stats_from_array in case input array is empty
            # this might occur in case of one cell mesh,
            # where analyzer.adj_ratio is empty.
            if metric_dict["array"].shape[0] == 0:
                continue
            _max, _min, _mean, _std = self.stats_from_array(metric_dict["array"])
            status = ""

            if "max" in metric_dict:
                status = "[red]Not good" if _max > metric_dict["max"] else "[green]Ok"

                if _max > metric_dict["max"]:
                    n_failed = np.count_nonzero(
                        metric_dict["array"] > metric_dict["max"]
                    )
                    self.concerns.append(
                        f"* Found {n_failed} elements with "
                        f"'{metric_name}' greater than "
                        f" max. value {metric_dict['max']}."
                    )

            if _max > 1e-4 or not metric_dict["sci_not"]:
                stats_table.add_row(
                    metric_name,
                    f"{_max:.4f}",
                    f"{_min:.4f}",
                    f"{_mean:.4f}",
                    f"{_std:.4f}",
                    status,
                )
            else:
                stats_table.add_row(
                    metric_name,
                    f"{_max:4e}",
                    f"{_min:.4e}",
                    f"{_mean:.4e}",
                    f"{_std:.4e}",
                )

        panel = Panel(
            stats_table,
            title="[yellow bold]Quality Stats.",
            title_align="left",
            expand=True,
        )

        self.console.print(panel)
        self.console.print()

    def report_file_size(self, filename: str):
        fsize = humanize.naturalsize(os.path.getsize(filename))
        self.console.print(
            f"Inspecting file [cyan]'{filename}'[/] (file size: {fsize})"
        )
        self.console.print()

    def report_bounding_box(self):
        self.console.print("[yellow bold]Mesh bounding box: ")
        for point in self.analyzer.bounding_box():
            print(
                f"\t{point}",
            )
        self.console.print()

    def report_concerns(self):
        if self.concerns:
            concerns_table = Table(box=None)
            concerns_table.add_column("", justify="left")

            for concern in self.concerns:
                concerns_table.add_row(f"[red]{concern}")

            concerns_table.add_row("")

            panel = Panel(
                concerns_table,
                expand=True,
                title="[yellow bold]Concerns",
                title_align="left",
            )
            self.console.print(panel)
            self.console.print()


class Reporter2D(Reporter):
    def __init__(self, console: Console, mesh: MeshReader, filename: str) -> None:
        super().__init__(console, mesh, filename)

    def report_elements_count(self):
        face_count = self.analyzer.n_faces
        edge_count = self.analyzer.n_edges
        point_count = self.analyzer.n_points

        tree = Tree(label="[yellow bold]Elements Stats.")

        tree.add("Mesh is 2-Dimensional")
        points_branch = tree.add(f"Points count = {point_count}", highlight=True)

        duplicate_nodes = self.analyzer.duplicate_nodes_count()
        duplicate_nodes_branch = (
            "[green](Ok)" if duplicate_nodes == 0 else "[red](Not good)"
        )

        if duplicate_nodes > 0:
            self.concerns.append(f"* Found {duplicate_nodes} duplicate nodes.")

        points_branch.add(
            f"Duplicate nodes = {duplicate_nodes} {duplicate_nodes_branch}"
        )

        tree.add(
            f"Edges count = {edge_count}"
            f" (including {self.analyzer.n_boundary_edges} boundary edges)"
        )
        faces_branch = tree.add(f"Faces count = {face_count}")

        if self.analyzer.n_quad > 0:
            quad_pct = (self.analyzer.n_quad / self.analyzer.n_faces) * 100
            faces_branch.add(f"Quadilaterals: {self.analyzer.n_quad} ({quad_pct:.1f}%)")
        if self.analyzer.n_tri > 0:
            tri_pct = (self.analyzer.n_tri / self.analyzer.n_faces) * 100
            faces_branch.add(f"Triangles: {self.analyzer.n_tri} ({tri_pct:.1f}%)")

        self.console.print(tree)

    def report(self, rules: Dict[str, float]):
        with self.console.status("Collecting cell types.."):
            self.analyzer = Analyzer2D(self.mesh)  # type: ignore
            self.analyzer.count_face_types()

        with self.console.status("Analyzing faces..."):
            self.analyzer.analyze_faces()

        with self.console.status("Checking non-orthogonality..."):
            self.analyzer.analyze_non_ortho()

        with self.console.status("Checking adjacent cells volume ratio..."):
            self.analyzer.analyze_adjacents_area_ratio()

        self.report_file_size(self.filename)

        if not rules:
            self.console.print("No quality rules file found, using defaults.\n")
        else:
            self.console.print(f"Found quality rules file: '{rules['fname']}'\n")

        self.report_bounding_box()
        self.report_elements_count()

        quality_metric_dict = {
            "Face Area": {
                "array": self.analyzer.face_areas,
                "sci_not": True,
            },
            "Face Aspect Ratio": {
                "array": self.analyzer.face_aspect_ratios,
                "sci_not": False,
                "max": rules.get("max_face_aspect_ratio", 20),
            },
            "Non-Orthogonality": {
                "array": self.analyzer.non_ortho,
                "sci_not": False,
                "max": rules.get("max_non_orhto", 60),
            },
            "Adjacent Faces Area Ratio": {
                "array": self.analyzer.adj_ratio,
                "sci_not": False,
                "max": rules.get("max_neighbor_area_ratio", 15),
            },
        }

        self.report_mesh_stats(quality_metric_dict)
        self.report_concerns()


class Reporter3D(Reporter):
    def __init__(self, console: Console, mesh: MeshReader, filename: str) -> None:
        super().__init__(console, mesh, filename)

    def report_elements_count(self):
        cell_count = self.analyzer.n_cells
        face_count = self.analyzer.n_faces
        point_count = self.analyzer.n_points

        tree = Tree(label="[yellow bold]Elements Stats.")

        tree.add("Mesh is 3-Dimensional")
        points_branch = tree.add(f"Points count = {point_count}", highlight=True)

        duplicate_nodes = self.analyzer.duplicate_nodes_count()
        duplicate_nodes_status = (
            "[green](Ok)" if duplicate_nodes == 0 else "[red](Error)"
        )
        points_branch.add(
            f"Duplicate nodes = {duplicate_nodes} {duplicate_nodes_status}"
        )
        faces_branch = tree.add(
            f"Faces count = {face_count}"
            f" (including {self.analyzer.n_boundary_faces} boundary faces)"
        )

        if self.analyzer.n_quad > 0:
            quad_pct = (self.analyzer.n_quad / self.analyzer.n_faces) * 100
            faces_branch.add(f"Quadilaterals: {self.analyzer.n_quad} ({quad_pct:.1f}%)")
        if self.analyzer.n_tri > 0:
            tri_pct = (self.analyzer.n_tri / self.analyzer.n_faces) * 100
            faces_branch.add(f"Triangles: {self.analyzer.n_tri} ({tri_pct:.1f}%)")

        cells_branch = tree.add(f"Cells count = {cell_count}")
        for ctype, count in (
            ("Hexahedron", self.analyzer.hex_count),
            ("Tetrahedron", self.analyzer.tetra_count),
            ("Wedge", self.analyzer.wedge_count),
            ("Pyramid", self.analyzer.pyramid_count),
        ):
            if count == 0:
                continue

            pct = (count / self.analyzer.n_cells) * 100
            cells_branch.add(f"{ctype}s: {count} ({pct:.1f}%)")

        self.console.print(tree)

    def report(self, rules: Dict[str, float]):
        with self.console.status("Collecting cell types.."):
            self.analyzer = Analyzer3D(self.mesh)  # type: ignore
            self.analyzer.count_cell_types()

        with self.console.status("Analyzing faces..."):
            self.analyzer.analyze_faces()

        with self.console.status("Analyzing cells..."):
            self.analyzer.analyze_cells()

        with self.console.status("Checking non-orthogonality..."):
            self.analyzer.analyze_non_ortho()

        with self.console.status("Checking adjacent cells volume ratio..."):
            self.analyzer.analyze_adjacents_volume_ratio()

        self.report_file_size(self.filename)

        if not rules:
            self.console.print("No quality rules file found, using defaults.\n")
        else:
            self.console.print(f"Found quality rules file: '{rules['fname']}'\n")

        self.report_bounding_box()
        self.report_elements_count()

        quality_metrics_dict = {
            "Face Area": {
                "array": self.analyzer.face_areas,
                "sci_not": True,
            },
            "Face Aspect Ratio": {
                "array": self.analyzer.face_aspect_ratios,
                "sci_not": False,
                "max": rules.get("max_face_aspect_ratio", 20),
            },
            "Cell Volume": {"array": self.analyzer.cells_volumes, "sci_not": True},
            "Non-Orthogonality": {
                "array": self.analyzer.non_ortho,
                "sci_not": False,
                "max": rules.get("max_non_orhto", 60),
            },
            "Adjacent Cells Volume Ratio": {
                "array": self.analyzer.adj_ratio,
                "sci_not": False,
                "max": rules.get("max_neighbor_volume_ratio", 15),
            },
        }

        self.report_mesh_stats(quality_metrics_dict)
        self.report_concerns()



"""neatmesh common constants and functions"""

from meshio import Mesh

meshio_3d = {
    "tetra",
    "tetra10",
    "hexahedron",
    "hexahedron20",
    "hexahedron24",
    "hexahedron27",
    "wedge",
    "wedge12",
    "wedge15",
    "pyramid",
    "pyramid13",
    "pyramid14",
}

meshio_type_to_alpha = {
    "vertex": "vertex",
    "line": "line",
    "line3": "line",
    "triangle": "triangle",
    "triangle6": "triangle",
    "triangle7": "triangle",
    "quad": "quad",
    "quad8": "quad",
    "quad9": "quad",
    "tetra": "tetra",
    "tetra10": "tetra",
    "hexahedron": "hexahedron",
    "hexahedron20": "hexahedron",
    "hexahedron24": "hexahedron",
    "hexahedron27": "hexahedron",
    "wedge": "wedge",
    "wedge12": "wedge",
    "wedge15": "wedge",
    "pyramid": "pyramid",
    "pyramid13": "pyramid",
    "pyramid14": "pyramid",
}

meshio_2d = {
    "triangle",
    "quad",
}

meshio_1d = {
    "vertex",
    "line",
    "line3",
}


def is_3d(mesh: Mesh) -> bool:
    """Check if a meshio mesh is 3-dimensional"""
    for cell_block in mesh.cells:
        # first 3D element type is enough.
        if meshio_type_to_alpha[cell_block.type] in meshio_3d:
            return True
    return False


def is_2d(mesh: Mesh) -> bool:
    """Check ifa meshio mesh is 2-dimensional"""
    if is_3d(mesh):
        return False

    for cell_block in mesh.cells:
        if meshio_type_to_alpha[cell_block.type] in meshio_2d:
            return True
    return False



"""neatmesh 2D & 3D mesh readers"""
# pylint: disable=too-few-public-methods, too-many-instance-attributes
from typing import Callable, Dict, FrozenSet, List, Set, Tuple

import meshio

from .common import is_2d, is_3d, meshio_1d, meshio_2d, meshio_3d, meshio_type_to_alpha
from .exceptions import InvalidMeshException, NonSupportedElement


# pylint: disable=unused-private-member
class MeshReader:
    """Base class for readers, do not use directly"""

    def __init__(self, mesh: meshio.Mesh) -> None:
        self.mesh = mesh
        self.points = self.mesh.points
        self.n_points = len(self.points)


class MeshReader2D(MeshReader):
    """A 2D mesh reader"""

    def __init__(self, mesh: meshio.Mesh) -> None:
        super().__init__(mesh)

        # list of points labels of processed edges
        self.edges: List[Tuple] = []
        self.__edges_sets: Set[FrozenSet] = set()
        self.edge_to_edgeid: Dict[FrozenSet, int] = {}

        # maps edge id to a list of faces id. An edge is shared by max. 2 faces.
        self.edgeid_to_faceid: Dict[int, List[int]] = {}

        # keep track of the edge id to be processed.
        self.__current_edgeid: int = 0

        # keep track of the face id to be processed.
        self.__current_faceid: int = 0

        self.__check_mesh()
        self.__process_mesh()

    def __check_mesh(self):
        """Look for 2D cell types, and check if mesh contains unsupported types"""
        self.n_faces = 0
        self.cell_blocks = []

        for cell_block in self.mesh.cells:
            ctype = meshio_type_to_alpha.get(cell_block.type, "unsupported")

            if ctype in meshio_2d and cell_block.data.size > 0:
                self.cell_blocks.append(cell_block)
                self.n_faces += len(cell_block.data)

            elif ctype not in meshio_2d and cell_block.type not in meshio_1d:
                raise NonSupportedElement(
                    f"neatmesh does not support element type: {cell_block.type}"
                )

        if not self.cell_blocks:
            raise InvalidMeshException("No 2D elements were found in mesh")

    def __process_mesh(self) -> None:
        """Get edges of each 2D cell, and assign owner/neighbor of each edge"""
        for cell_block in self.cell_blocks:
            faces = cell_block.data
            face_type = meshio_type_to_alpha[cell_block.type]

            if face_type == "quad":
                edges_func = quad_face_edges
            else:
                edges_func = tri_face_edges

            for face in faces:
                edges = edges_func(face)
                for edge in edges:
                    f_edge = frozenset(edge)
                    # have we met `edge` before?
                    if f_edge not in self.__edges_sets:
                        self.edge_to_edgeid[f_edge] = self.__current_edgeid
                        self.edges.append(edge)
                        self.__edges_sets.add(f_edge)

                        self.edgeid_to_faceid[self.__current_edgeid] = [
                            self.__current_faceid,
                            -1,
                        ]
                        self.__current_edgeid += 1
                    else:
                        # link edge to face who owns it
                        edge_id = self.edge_to_edgeid[f_edge]
                        self.edgeid_to_faceid[edge_id][1] = self.__current_faceid

                self.__current_faceid += 1


class MeshReader3D(MeshReader):
    """A 3D mesh reader"""

    def __init__(self, mesh: meshio.Mesh) -> None:
        super().__init__(mesh)

        # list of points labels of processed faces (all types)
        self.faces: List[Tuple[int, ...]] = []

        self.__faces_set: Set[FrozenSet] = set()

        # map face points to face index in `faces`
        self.face_to_faceid: Dict[FrozenSet, int] = {}

        # maps face id to a list of cells id. A face is shared by max. 2 cells.
        self.faceid_to_cellid: Dict[int, List[int]] = {}

        # keep track of the face id to be processed.
        self.__current_faceid: int = 0

        # keep track of the cell id to be processed.
        self.__current_cellid: int = 0

        self.__check_mesh()
        self.__process_mesh()

    def __check_mesh(self):
        """Look for 2D cell types, and check if mesh contains unsupported types"""
        self.n_cells = 0
        self.cell_blocks = []

        for cell_block in self.mesh.cells:
            ctype = meshio_type_to_alpha.get(cell_block.type, "unsupported")

            if ctype in meshio_3d and cell_block.data.size > 0:
                self.cell_blocks.append(cell_block)
                self.n_cells += len(cell_block.data)

            elif ctype not in meshio_2d and cell_block.type not in meshio_1d:
                raise NonSupportedElement(
                    f"neatmesh does not support element type: {cell_block.type}"
                )

        if not self.cell_blocks:
            raise InvalidMeshException("No 3D elements were found in mesh")

    def __process_mesh(self) -> None:
        """Get faces of each 3D cell, and assign owner/neighbor of each face"""
        for cell_block in self.cell_blocks:
            cells = cell_block.data

            for cell in cells:
                cell_type = meshio_type_to_alpha[cell_block.type]
                faces_func = cell_type_to_faces_func[cell_type]
                faces = faces_func(cell)
                for face in faces:
                    fface = frozenset(face)
                    # have we met `face` before?
                    if fface not in self.__faces_set:
                        self.face_to_faceid[fface] = self.__current_faceid
                        self.__faces_set.add(fface)

                        # add face points labels to `faces`
                        self.faces.append(face)
                        self.faceid_to_cellid[self.__current_faceid] = [
                            self.__current_cellid,
                            -1,
                        ]
                        self.__current_faceid += 1
                    else:
                        # link the face to the cell who owns it
                        face_id = self.face_to_faceid[fface]
                        self.faceid_to_cellid[face_id][1] = self.__current_cellid

                self.__current_cellid += 1


def hex_cell_faces(cell: List) -> Tuple[Tuple[int, ...], ...]:
    """Returns coordinates of 6 faces of a hexahedron cell, using meshio nodes ordering
    Args:
        cell (List): list of points defining the cell
    Returns:
        List[List]: list of list of faces points labels
    """
    return (
        (cell[1], cell[2], cell[6], cell[5]),
        (cell[0], cell[4], cell[7], cell[3]),
        (cell[3], cell[7], cell[6], cell[2]),
        (cell[0], cell[1], cell[5], cell[4]),
        (cell[4], cell[5], cell[6], cell[7]),
        (cell[0], cell[3], cell[2], cell[1]),
    )


def wedge_cell_faces(cell: List) -> Tuple[Tuple[int, ...], ...]:
    """Returns coordinates of 5 faces of a wedge cell,
    using meshio nodes ordering for wedge
    Args:
        cell (List): list of points defining the cell
    Returns:
        List[List]: list of list of faces points labels
    """
    return (
        (cell[0], cell[2], cell[1], -1),
        (cell[3], cell[4], cell[5], -1),
        (cell[3], cell[0], cell[1], cell[4]),
        (cell[0], cell[3], cell[5], cell[2]),
        (cell[1], cell[2], cell[5], cell[4]),
    )


def tetra_cell_faces(cell: List) -> Tuple[Tuple[int, ...], ...]:
    """Returns coordinates of 4 faces of a tetrahedral cell,
    using meshio nodes ordering for tetra
    Args:
        cell (List): list of points defining the cell
    Returns:
        List[List]: list of list of faces points labels
    """
    return (
        (cell[0], cell[2], cell[1], -1),
        (cell[1], cell[2], cell[3], -1),
        (cell[0], cell[1], cell[3], -1),
        (cell[0], cell[3], cell[2], -1),
    )


def pyramid_cell_faces(cell: List) -> Tuple[Tuple[int, ...], ...]:
    """Returns coordinates of 4 faces of a tetrahedral cell,
    using meshio nodes ordering for pyramid
    Args:
        cell (List): list of points defining the cell
    Returns:
        List[List]: list of list of faces points labels
    """
    return (
        (cell[2], cell[1], cell[0], cell[3]),
        (cell[2], cell[3], cell[4], -1),
        (cell[1], cell[4], cell[0], -1),
        (cell[3], cell[0], cell[4], -1),
    )


def quad_face_edges(face: List) -> Tuple[Tuple[int, ...], ...]:
    """Get quadilateral face edges labels"""
    return (
        (face[0], face[1]),
        (face[1], face[2]),
        (face[2], face[3]),
        (face[3], face[0]),
    )


def tri_face_edges(face: List) -> Tuple[Tuple[int, ...], ...]:
    """Get triangle face edges labels"""
    return (
        (face[0], face[1]),
        (face[1], face[2]),
        (face[2], face[0]),
    )


cell_type_to_faces_func: Dict[str, Callable] = {
    "hexahedron": hex_cell_faces,
    "tetra": tetra_cell_faces,
    "wedge": wedge_cell_faces,
    "pyramid": pyramid_cell_faces,
}


def assign_reader(mesh_file_path: str) -> MeshReader:
    """Assign a 2D or 3D MeshReader, given the mesh file.

    Args:
        mesh_file_path (str): Path to mesh file

    Raises:
        InvalidMeshException: When assign_reader cannot decide mesh dimensions

    Returns:
        MeshReader: 2D or 3D Reader
    """
    try:
        mesh = meshio.read(mesh_file_path)
    except meshio.ReadError as exception:
        error = "Could not read mesh file (meshio error).\n"
        error += f"{exception}"
        raise InvalidMeshException(error) from exception

    if is_3d(mesh):
        return MeshReader3D(mesh)
    if is_2d(mesh):
        return MeshReader2D(mesh)

    raise InvalidMeshException("Couldn't decide on mesh dimensionality")



"""neatmesh mesh 2D & 3D analyzers"""
# pylint: disable=too-many-instance-attributes
from typing import Tuple

import numpy as np

from .common import meshio_type_to_alpha
from .geometry import (
    dot,
    dot_normalize,
    hex_data_from_tensor,
    pyramid_data_from_tensor,
    quad_data_from_tensor,
    tetra_data_from_tensor,
    tri_data_from_tensor,
    wedge_data_from_tensor,
)
from .reader import MeshReader2D, MeshReader3D


class Analyzer2D:
    """Analyze a 2D mesh and provide edges and faces count,
    faces areas, faces aspect ratios and adjacent faces volume ratios.

    Analyzer2D requires an instance of a MeshReader2D as an input.
    """

    def __init__(self, reader: MeshReader2D) -> None:
        self.reader = reader

        self.points = self.reader.points
        self.n_points = reader.n_points

        self.n_edges = len(reader.edges)
        self.n_faces = reader.n_faces

        self.n_quad = 0
        self.n_tri = 0

        self.face_centers: np.ndarray = np.array([]).reshape(0, 3)
        self.face_areas: np.ndarray = np.array([])
        self.face_aspect_ratios: np.ndarray = np.array([])

        # Add 3rd dimension to points,
        # to use geometry module 3D tri & quad functions
        if self.points.shape[1] == 2:
            self.__3d_points = np.concatenate(
                [self.points, np.zeros(shape=(self.n_points, 1))], axis=1
            )
        else:
            self.__3d_points = self.points

        # This translates list of edge tuples, to point coordinates
        self.__edges_tensor = np.take(self.__3d_points, self.reader.edges, axis=0)

        self.__interior_edges = np.array([])
        self.n_boundary_edges: int = 0

        self.non_ortho = np.array([])
        self.adj_ratio = np.array([])

    def duplicate_nodes_count(self) -> int:
        """Count number of duplicate points/nodes"""
        return self.points.shape[0] - np.unique(self.points, axis=0).shape[0]

    def bounding_box(self) -> Tuple:
        """Return coordinates of bounding box of the mesh (min & max coords)"""
        x_min, y_min = np.min(self.points, axis=0)
        x_max, y_max = np.max(self.points, axis=0)

        return (
            (x_min, y_min),
            (x_max, y_max),
        )

    def count_face_types(self) -> None:
        """Count and update the number of quad and triangle faces"""
        for cell_block in self.reader.cell_blocks:
            alpha_face_type = meshio_type_to_alpha[cell_block.type]
            if alpha_face_type == "quad":
                self.n_quad += len(cell_block.data)

            elif alpha_face_type == "triangle":
                self.n_tri += len(cell_block.data)

    def analyze_faces(self) -> None:
        """Call appropriate methods for each face type and update the following:
        faces centers, faces areas and face aspect ratios.
        """
        for cell_block in self.reader.cell_blocks:
            face_type = meshio_type_to_alpha[cell_block.type]
            faces_tensor = np.take(self.__3d_points, cell_block.data, axis=0)

            if face_type == "quad":
                centers, _, areas, aspect_ratios = quad_data_from_tensor(faces_tensor)
            else:
                centers, _, areas, aspect_ratios = tri_data_from_tensor(faces_tensor)

            self.face_centers = np.concatenate([self.face_centers, centers])
            self.face_areas = np.concatenate([self.face_areas, areas])
            self.face_aspect_ratios = np.concatenate(
                [self.face_aspect_ratios, aspect_ratios]
            )

    def analyze_non_ortho(self) -> None:
        """For each internal edge, calculate the non-orthogonality between the faces.
        Non-orthogonality is defined as the angle (in degrees) between the vector
        connecting two neighbor faces centroids, and shared edge normal vector.
        """
        # owner_neighbor is 2D matrix of shape (n_interior_edges, 2)
        # first column is the index of the owner faces,
        # second column is the index of neighbor faces.
        # second column contains '-1' for boudnary edges.
        owner_neighbor = np.asarray(list(self.reader.edgeid_to_faceid.values()))

        # filter our boundary edges
        interior_edges_mask = owner_neighbor[:, 1] != -1

        # We will need interior_edges later in adjacent cells volume ratio
        self.__interior_edges = owner_neighbor[interior_edges_mask]

        self.n_boundary_edges = self.n_edges - self.__interior_edges.shape[0]

        owners, neighbors = self.__interior_edges[:, 0], self.__interior_edges[:, 1]
        owner_centers = np.take(self.face_centers, owners, axis=0)
        neighbor_centers = np.take(self.face_centers, neighbors, axis=0)

        # internal_edges_tensor has shape (n_internal_edges, 2, 3)
        # and represents the coordinates of first and second coordinates of
        # each edge.
        internal_edges_tensor = self.__edges_tensor[interior_edges_mask]
        edges_vectors = internal_edges_tensor[:, 1, :] - internal_edges_tensor[:, 0, :]

        # array of vectors connecting adjacent faces centroids.
        neigbor_owner_vectors = neighbor_centers - owner_centers

        costheta = np.abs(dot_normalize(neigbor_owner_vectors, edges_vectors))
        self.non_ortho = 90 - (np.arccos(costheta) * (180.0 / np.pi))

    def analyze_adjacents_area_ratio(self) -> None:
        """Calculate the area ratio for each two neighbor faces (max/min)"""
        adjacent_faces_areas = np.take(self.face_areas, self.__interior_edges, axis=0)
        self.adj_ratio = np.max(
            [
                adjacent_faces_areas[:, 0] / adjacent_faces_areas[:, 1],
                adjacent_faces_areas[:, 1] / adjacent_faces_areas[:, 0],
            ],
            axis=0,
        )


class Analyzer3D:
    """Analyze a 3D mesh and provide faces and cells count,
    faces areas, faces aspect ratios and adjacent cells volume ratios.

    Analyzer3D requires an instance of a MeshReader3D as an input.
    """

    def __init__(self, reader: MeshReader3D) -> None:
        self.reader = reader

        self.points = self.reader.points
        self.n_points = reader.n_points

        self.faces = np.asarray(self.reader.faces)
        self.n_faces = len(reader.faces)

        self.n_cells = reader.n_cells

        self.hex_count = 0
        self.tetra_count = 0
        self.wedge_count = 0
        self.pyramid_count = 0
        self.has_tri = False
        self.has_quad = False

        self.face_centers = np.zeros(shape=(self.n_faces, 3))
        self.face_normals = np.zeros(shape=(self.n_faces, 3))
        self.face_areas = np.zeros(shape=(self.n_faces,))
        self.face_aspect_ratios = np.zeros(shape=(self.n_faces,))

        self.n_tri = 0
        self.n_quad = 0

        self.cells_centers: np.ndarray = np.array([]).reshape(0, 3)
        self.cells_volumes: np.ndarray = np.array([])

        self.interior_faces: np.ndarray = np.array([])
        self.n_boundary_faces: int = 0

        self.non_ortho: np.ndarray = np.array([])
        self.adj_ratio: np.ndarray = np.array([])

    def duplicate_nodes_count(self) -> int:
        """Count number of duplicate points/nodes"""
        return self.points.shape[0] - np.unique(self.points, axis=0).shape[0]

    def bounding_box(self) -> Tuple:
        """Return coordinates of bounding box of the mesh (min & max coords)"""
        x_min, y_min, z_min = np.min(self.points, axis=0)
        x_max, y_max, z_max = np.max(self.points, axis=0)

        return (
            (x_min, y_min, z_min),
            (x_max, y_max, z_max),
        )

    def count_cell_types(self) -> None:
        """Count and update the number of cells for each cell type"""
        for cell_block in self.reader.cell_blocks:
            alpha_cell_type = meshio_type_to_alpha[cell_block.type]
            if alpha_cell_type == "hexahedron":
                self.hex_count += len(cell_block.data)
                self.has_quad = True

            elif alpha_cell_type == "tetra":
                self.tetra_count += len(cell_block.data)
                self.has_tri = True

            elif alpha_cell_type == "pyramid":
                self.pyramid_count += len(cell_block.data)
                self.has_quad = True
                self.has_tri = True

            elif alpha_cell_type == "wedge":
                self.wedge_count += len(cell_block.data)
                self.has_quad = True
                self.has_tri = True

    def analyze_faces(self):
        """Call appropriate methods for each face type and update the following:
        faces centers, faces areas, face normals and face aspect ratios.
        """
        if self.has_tri:
            tri_mask = self.faces[:, -1] == -1
            tri_faces = self.faces[tri_mask][:, :-1]
            tri_faces_tensor = np.take(self.points, tri_faces, axis=0)[:, 0:3, :]

            (
                self.face_centers[tri_mask],
                self.face_normals[tri_mask],
                self.face_areas[tri_mask],
                self.face_aspect_ratios[tri_mask],
            ) = tri_data_from_tensor(tri_faces_tensor)
            self.n_tri += len(tri_faces)

        if self.has_quad:
            quad_mask = self.faces[:, -1] != -1
            quad_faces = self.faces[quad_mask]

            quad_faces_tensor = np.take(self.points, quad_faces, axis=0)[:, 0:5, :]
            (
                self.face_centers[quad_mask],
                self.face_normals[quad_mask],
                self.face_areas[quad_mask],
                self.face_aspect_ratios[quad_mask],
            ) = quad_data_from_tensor(quad_faces_tensor)
            self.n_quad += len(quad_faces)

    def analyze_cells(self) -> None:
        """Call appropriate methods for each cell type and update cells centers
        and cells volumes.
        """
        cell_type_handler_map = {
            "hexahedron": hex_data_from_tensor,
            "tetra": tetra_data_from_tensor,
            "wedge": wedge_data_from_tensor,
            "pyramid": pyramid_data_from_tensor,
        }

        for cell_block in self.reader.cell_blocks:
            ctype = meshio_type_to_alpha[cell_block.type]
            handler = cell_type_handler_map[ctype]
            data_tensor = np.take(self.points, cell_block.data, axis=0)
            centers, vols = handler(data_tensor)
            self.cells_centers = np.concatenate([self.cells_centers, centers], axis=0)
            self.cells_volumes = np.concatenate([self.cells_volumes, vols], axis=0)

    def analyze_non_ortho(self) -> None:
        """For each internal face, calculate the non-orthogonality between the cells.
        Non-orthogonality is defined as the angle (in degrees) between the vector
        connecting two neighbor cells centroids, and shared face normal vector.
        """
        owner_neighbor = np.asarray(list(self.reader.faceid_to_cellid.values()))
        interior_faces_mask = owner_neighbor[:, 1] != -1

        # We will need interior_faces later in adjacent cells volume ratio
        self.interior_faces = owner_neighbor[interior_faces_mask]

        self.n_boundary_faces = self.n_faces - self.interior_faces.shape[0]

        owners, neighbors = self.interior_faces[:, 0], self.interior_faces[:, 1]
        owner_centers = np.take(self.cells_centers, owners, axis=0)
        neighbor_centers = np.take(self.cells_centers, neighbors, axis=0)

        interior_face_normals = self.face_normals[interior_faces_mask]
        nei_owner_vectors = neighbor_centers - owner_centers
        dot_product = dot(nei_owner_vectors, interior_face_normals)

        nei_owner_vectors[dot_product < 0] = -nei_owner_vectors[dot_product < 0]

        costheta = dot_normalize(nei_owner_vectors, interior_face_normals)
        self.non_ortho = np.arccos(costheta) * (180.0 / np.pi)

    def analyze_adjacents_volume_ratio(self) -> None:
        """Calculate the area ratio for each two neighbor cells (max/min)"""
        adjacent_cells_vol = np.take(self.cells_volumes, self.interior_faces, axis=0)
        self.adj_ratio = np.max(
            [
                adjacent_cells_vol[:, 0] / adjacent_cells_vol[:, 1],
                adjacent_cells_vol[:, 1] / adjacent_cells_vol[:, 0],
            ],
            axis=0,
        )



"""geometry related functions for faces and cells"""
# pylint: disable=invalid-name, too-many-locals
from typing import Tuple

import numpy as np
from numpy.linalg import det, norm


def tri_data_from_tensor(tri_faces_tensor: np.ndarray) -> Tuple[np.ndarray, ...]:
    """Calculate triangular faces centers, normal, area and aspect ratios.

    Args:
        tri_faces_tensor (np.ndarray): Triangle faces tensor
        shape of `tri_faces_tensor` is (n_faces, 3 (no. points), 3)

    Returns:
        Tuple[np.ndarray, ...]: centers, normals, areas and aspect ratios.
    """
    tri_centers = np.mean(tri_faces_tensor, axis=1)
    tri_normals = np.cross(
        tri_faces_tensor[:, 1, :] - tri_faces_tensor[:, 0, :],
        tri_faces_tensor[:, 2, :] - tri_faces_tensor[:, 0, :],
    )
    tri_areas = norm(tri_normals, axis=1) / 2.0
    tri_edges_norms = np.array(
        [
            norm(tri_faces_tensor[:, 0, :] - tri_faces_tensor[:, 1, :], axis=1),
            norm(tri_faces_tensor[:, 0, :] - tri_faces_tensor[:, 2, :], axis=1),
            norm(tri_faces_tensor[:, 1, :] - tri_faces_tensor[:, 2, :], axis=1),
        ]
    ).swapaxes(0, 1)

    tri_aspect_ratios = np.max(tri_edges_norms, axis=1) / np.min(
        tri_edges_norms, axis=1
    )
    return tri_centers, tri_normals, tri_areas, tri_aspect_ratios


def quad_data_from_tensor(quad_faces_tensor: np.ndarray) -> Tuple[np.ndarray, ...]:
    """Calculate quad faces centers, normal, area and aspect ratios.

    Args:
        quad_faces_tensor (np.ndarray): Quad faces tensor
        shape of `quad_faces_tensor` is (n_faces, 4 (no. points), 3)

    Returns:
        Tuple[np.ndarray, ...]: centers, normals, areas and aspect ratios.
    """
    # Quads geometrics centers
    gc = np.mean(quad_faces_tensor, axis=1)

    # Quad sub-triangles
    edges = ((0, 1), (1, 2), (2, 3), (3, 0))

    sub_triangles_centroids_unswapped = []
    sub_triangles_normals_unswapped = []

    for edge in edges:
        sub_triangles_centroids_unswapped.append(
            np.mean(
                [
                    gc,
                    quad_faces_tensor[:, edge[0], :],
                    quad_faces_tensor[:, edge[1], :],
                ],
                axis=0,
            )
        )

        sub_triangles_normals_unswapped.append(
            np.cross(
                gc - quad_faces_tensor[:, edge[0], :],
                quad_faces_tensor[:, edge[1], :] - quad_faces_tensor[:, edge[0], :],
                axis=1,
            )
        )

    sub_triangles_centroids = np.swapaxes(
        np.asarray(sub_triangles_centroids_unswapped), 0, 1
    )
    sub_triangles_normals = np.swapaxes(
        np.asarray(sub_triangles_normals_unswapped), 0, 1
    )

    sub_triangles_areas = norm(sub_triangles_normals, axis=2) / 2.0
    area_weighted_centroids = (
        sub_triangles_areas[:, :, np.newaxis] * sub_triangles_centroids
    )
    quad_centroids = np.sum(area_weighted_centroids, axis=1)
    quad_areas = np.sum(sub_triangles_areas, axis=1)[:, np.newaxis]
    quad_centroids /= quad_areas

    quad_normals = np.sum(sub_triangles_normals, axis=1)

    quad_edges_norms = np.array(
        [
            norm(quad_faces_tensor[:, 0, :] - quad_faces_tensor[:, 1, :], axis=1),
            norm(quad_faces_tensor[:, 1, :] - quad_faces_tensor[:, 2, :], axis=1),
            norm(quad_faces_tensor[:, 2, :] - quad_faces_tensor[:, 3, :], axis=1),
        ]
    )
    quad_aspect_ratios = np.max(quad_edges_norms, axis=0) / np.min(
        quad_edges_norms, axis=0
    )

    return quad_centroids, quad_normals, quad_areas.flatten(), quad_aspect_ratios


def tetra_data_from_tensor(
    tetra_cells_tensor: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate tetrahedron cells centers and volumes

    Args:
        tetra_cells_tensor (np.ndarray): Tetrahedron cells tensor
        shape of `tetra_cells_tensor` is (n_cells, 4 (no. points), 3)

    Returns:
        Tuple[np.ndarray, ...]: centers and volumes
    """
    tetra_centers = np.mean(tetra_cells_tensor, axis=1)
    tetra_vols = (
        np.abs(
            det(
                np.array(
                    (
                        tetra_cells_tensor[:, 0, :] - tetra_cells_tensor[:, 1, :],
                        tetra_cells_tensor[:, 1, :] - tetra_cells_tensor[:, 2, :],
                        tetra_cells_tensor[:, 2, :] - tetra_cells_tensor[:, 3, :],
                    )
                ).swapaxes(0, 1)
            )
        )
        / 6.0
    )
    return tetra_centers, tetra_vols


def pyramid_data_from_tensor(pyr_cells_tensor: np.ndarray) -> Tuple[np.ndarray, ...]:
    """Calculate pyramid cells centers and volumes

    Args:
        pyr_cells_tensor (np.ndarray): Pyramid cells tensor
        shape of `pyr_cells_tensor` is (n_cells, 4 (no. points), 3)

    Returns:
        Tuple[np.ndarray, ...]: centers and volumes
    """
    # Pyramid base area and centroid
    quad_base_tensor = pyr_cells_tensor[:, 0:-1, :]
    quad_centroids, quad_normals, quad_areas, _ = quad_data_from_tensor(
        quad_base_tensor
    )

    pyramids_apex = pyr_cells_tensor[:, -1, :]
    # pyramids_heights = norm(pyramids_apex - quad_centroids, axis=1)[:, np.newaxis]
    normals_unit_vecs = quad_normals / norm(quad_normals, axis=1)[:, np.newaxis]

    pyramids_heights = np.sum(
        (pyramids_apex - quad_centroids) * normals_unit_vecs, axis=1
    )

    pyramids_vol = (1.0 / 3.0) * quad_areas.flatten() * np.abs(pyramids_heights)
    pyramids_centroids = (0.75 * quad_centroids) + (0.25 * pyramids_apex)

    return pyramids_centroids, pyramids_vol


def wedge_data_from_tensor(wedge_cells_tensor: np.ndarray) -> Tuple[np.ndarray, ...]:
    """Calculate wedge cells centers and volumes

    Args:
        wedge_data_from_tensor (np.ndarray): Wedge cells tensor
        shape of `wedge_data_from_tensor` is (n_cells, 6 (no. points), 3)

    Returns:
        Tuple[np.ndarray, ...]: centers and volumes
    """
    gc = np.mean(wedge_cells_tensor, axis=1)[:, np.newaxis, :]
    n_wedge = wedge_cells_tensor.shape[0]
    wedges_vol = np.zeros(shape=(n_wedge, 1))

    # Divide wedge into 5 surfaces (two triangles and three quads)
    upper_tri_tensor = wedge_cells_tensor[:, [3, 4, 5], :]
    lower_tri_tensor = wedge_cells_tensor[:, [0, 1, 2], :]
    quad1_tensor = wedge_cells_tensor[:, [0, 1, 4, 3], :]
    quad2_tensor = wedge_cells_tensor[:, [1, 2, 5, 4], :]
    quad3_tensor = wedge_cells_tensor[:, [0, 3, 5, 2], :]

    # Upper & lower tetrahedrons
    upper_tetra_tensor = np.concatenate([upper_tri_tensor, gc], axis=1)
    lower_tetra_tensor = np.concatenate([lower_tri_tensor, gc], axis=1)
    tetras_tensor = np.concatenate([upper_tetra_tensor, lower_tetra_tensor], axis=0)

    tetras_data = tetra_data_from_tensor(tetras_tensor)

    upper_tetra_vol, lower_tetra_vol = (
        tetras_data[1][0:n_wedge],
        tetras_data[1][n_wedge:],
    )
    (upper_tetra_centers, lower_tetra_centers) = (
        tetras_data[0][0:n_wedge],
        tetras_data[0][n_wedge:],
    )

    wedges_vol = upper_tetra_vol[:, np.newaxis] + lower_tetra_vol[:, np.newaxis]

    # Three middle pyramids
    pyramid1_tensor = np.concatenate([quad1_tensor, gc], axis=1)
    pyramid2_tensor = np.concatenate([quad2_tensor, gc], axis=1)
    pyramid3_tensor = np.concatenate([quad3_tensor, gc], axis=1)
    pyramids_tensor = np.concatenate(
        [pyramid1_tensor, pyramid2_tensor, pyramid3_tensor], axis=0
    )

    pyramids_data = pyramid_data_from_tensor(pyramids_tensor)
    pyramid1_vol, pyramid2_vol, pyramid3_vol = (
        pyramids_data[1][0:n_wedge],
        pyramids_data[1][n_wedge : 2 * n_wedge],
        pyramids_data[1][2 * n_wedge :],
    )
    pyramid1_centers, pyramid2_centers, pyramid3_centers = (
        pyramids_data[0][0:n_wedge],
        pyramids_data[0][n_wedge : 2 * n_wedge],
        pyramids_data[0][2 * n_wedge :],
    )

    # Wedge volume
    wedges_vol += np.sum([pyramid1_vol, pyramid2_vol, pyramid3_vol], axis=0)[
        :, np.newaxis
    ]

    # Wedge centroid
    wedges_center = upper_tetra_centers * upper_tetra_vol[:, np.newaxis]
    wedges_center += lower_tetra_centers * lower_tetra_vol[:, np.newaxis]
    wedges_center += pyramid1_centers * pyramid1_vol[:, np.newaxis]
    wedges_center += pyramid2_centers * pyramid2_vol[:, np.newaxis]
    wedges_center += pyramid3_centers * pyramid3_vol[:, np.newaxis]
    wedges_center /= wedges_vol

    return wedges_center, wedges_vol.flatten()


def hex_data_from_tensor(hex_cells_tensor: np.ndarray) -> Tuple[np.ndarray, ...]:
    """Calculate hexahedron cells centers and volumes

    Args:
        hex_cells_tensor (np.ndarray): Hexahedron cells tensor
        shape of `hex_cells_tensor` is (n_cells, 8 (no. points), 3)

    Returns:
        Tuple[np.ndarray, ...]: centers and volumes
    """
    x = norm(hex_cells_tensor[:, 0, :] - hex_cells_tensor[:, 1, :], axis=1)
    y = norm(hex_cells_tensor[:, 0, :] - hex_cells_tensor[:, 3, :], axis=1)
    z = norm(hex_cells_tensor[:, 0, :] - hex_cells_tensor[:, 4, :], axis=1)

    hex_centers = np.mean(hex_cells_tensor, axis=1)
    hex_vols = x * y * z

    return hex_centers, hex_vols


def dot_normalize(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Element wise dot product with normalization"""
    return np.sum(x * y, axis=1) / ((norm(x, axis=1) * norm(y, axis=1)))


def dot(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Element wise dot product"""
    return np.sum(x * y, axis=1)



"""neatmesh cli entry point"""
import argparse
import os
import sys
from pathlib import Path

import toml
from rich.console import Console

from neatmesh.reader import MeshReader3D, assign_reader
from neatmesh.reporter import Reporter2D, Reporter3D


# pylint: disable=import-outside-toplevel
def version() -> str:
    """get neatmesh version"""
    try:
        from importlib import metadata

        __version__ = metadata.version("neatmesh")
    except ImportError:
        __version__ = "0.0.0"

    return __version__


def header_str():
    """neatmesh logo, version and license header"""
    return rf"""
                     __                      __
   ____  ___  ____ _/ /_____ ___  ___  _____/ /_
  / __ \/ _ \/ __ `/ __/ __ `__ \/ _ \/ ___/ __ \   Version: {version()}
 / / / /  __/ /_/ / /_/ / / / / /  __(__  ) / / /   License: MIT
/_/ /_/\___/\__,_/\__/_/ /_/ /_/\___/____/_/ /_/

"""


def error(msg: str) -> None:
    """display fatal error message and exit"""
    Console(stderr=True).print(f"[red][bold]Error:[/bold] {msg}[/red]")
    sys.exit(-1)


def main():
    """cli entry points"""
    print(header_str())

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_file", metavar="input", type=str, nargs=1, help="Input mesh file name"
    )

    args = parser.parse_args()
    filename = args.input_file[0]

    if not os.path.isfile(filename):
        error(f"file {filename} does not exist!")

    console = Console()

    with console.status("Reading mesh..."):
        mesh = assign_reader(filename)

    if isinstance(mesh, MeshReader3D):
        reporter = Reporter3D(console, mesh, filename)
    else:
        reporter = Reporter2D(console, mesh, filename)

    # Check quality rules file, if exists
    fname_stripped = Path(filename).stem
    quality_rules_candidates = [
        f"{fname_stripped}.toml",
        "neatmesh.toml",
        "quality.toml",
    ]

    rules_dict = {}
    for fname in quality_rules_candidates:
        if os.path.isfile(fname):
            rules_dict = toml.load(fname)
            rules_dict["fname"] = fname
            break

    reporter.report(rules_dict)



"""Neatmesh derived exceptions"""


class InvalidMeshException(Exception):
    """Raised by MeshReader in cases of not finding relevant cell types in
    given mesh or when encountering issues with mesh dimensionality"""

    ...


class NonSupportedElement(Exception):
    """Raised by Mesh Reader in case of encountering unsupported cell type"""

    ...
